"""API for implementing LynxKite operations."""

from __future__ import annotations

import asyncio
import enum
import functools
import json
import importlib.util
import inspect
import pathlib
import pkgutil
import subprocess
import traceback
import types
import typing
from dataclasses import dataclass

import pydantic

if typing.TYPE_CHECKING:
    from . import workspace

Catalog = dict[str, "Op"]
Catalogs = dict[str, Catalog]
CATALOGS: Catalogs = {}
EXECUTORS = {}

typeof = type  # We have some arguments called "type".

CACHE_WRAPPER = None  # Overwrite this to configure a caching mechanism.


def cached(func):
    if CACHE_WRAPPER is None:
        return func
    return CACHE_WRAPPER(func)


def type_to_json(t):
    if isinstance(t, type) and issubclass(t, enum.Enum):
        return {"enum": list(t.__members__.values())}
    if getattr(t, "__metadata__", None):
        return t.__metadata__[-1]
    return {"type": str(t)}


Type = typing.Annotated[typing.Any, pydantic.PlainSerializer(type_to_json, return_type=dict)]
LongStr = typing.Annotated[str, {"format": "textarea"}]
"""LongStr is a string type for parameters that will be displayed as a multiline text area in the UI."""
PathStr = typing.Annotated[str, {"format": "path"}]
# https://github.com/python/typing/issues/182#issuecomment-1320974824
ReadOnlyJSON: typing.TypeAlias = (
    typing.Mapping[str, "ReadOnlyJSON"]
    | typing.Sequence["ReadOnlyJSON"]
    | str
    | int
    | float
    | bool
    | None
)


class BaseConfig(pydantic.BaseModel):
    model_config = pydantic.ConfigDict()


class Parameter(BaseConfig):
    """Defines a parameter for an operation."""

    name: str
    default: typing.Any
    type: Type = None

    @staticmethod
    def options(name, options: list[str], default=None):
        e = enum.StrEnum(f"OptionsFor_{name}", [(o, o) for o in options])
        return Parameter.basic(name, default or options[0], e)

    @staticmethod
    def basic(name, default=None, type=None):
        if default is inspect._empty:
            default = None
        if type is None or type is inspect._empty:
            type = typeof(default) if default is not None else None
        return Parameter(name=name, default=default, type=type)


class ParameterGroup(BaseConfig):
    """Defines a group of parameters for an operation."""

    name: str
    selector: Parameter
    default: typing.Any
    groups: dict[str, list[Parameter]]
    type: str = "group"


class Position(enum.StrEnum):
    """Defines the position of an input or output in the UI."""

    LEFT = "left"
    RIGHT = "right"
    TOP = "top"
    BOTTOM = "bottom"

    def is_vertical(self):
        return self in (self.TOP, self.BOTTOM)

    @staticmethod
    def from_dir(dir: str) -> tuple[Position, Position]:
        """Returns the input and output positions based on the direction."""
        if dir == "left-to-right":
            return Position.LEFT, Position.RIGHT
        elif dir == "right-to-left":
            return Position.RIGHT, Position.LEFT
        elif dir == "top-to-bottom":
            return Position.TOP, Position.BOTTOM
        elif dir == "bottom-to-top":
            return Position.BOTTOM, Position.TOP
        else:
            raise ValueError(f"Invalid direction: {dir}")


class Input(BaseConfig):
    name: str
    type: Type
    position: Position


class Output(BaseConfig):
    name: str
    type: Type
    position: Position


@dataclass
class Result:
    """Represents the result of an operation.

    The `output` attribute is what will be used as input for other operations.
    The `display` attribute is used to send data to display on the UI. The value has to be
    JSON-serializable.
    `input_metadata` is a list of JSON objects describing each input.
    """

    output: typing.Any | None = None
    display: ReadOnlyJSON | None = None
    error: str | None = None
    input_metadata: list[dict[str, ReadOnlyJSON]] | None = None


def get_optional_type(type):
    """For a type like `int | None`, returns `int`. Returns `None` otherwise."""
    if isinstance(type, types.UnionType):
        match type.__args__:
            case (types.NoneType, type):
                return type
            case (type, types.NoneType):
                return type


def _param_to_type(name, value, type):
    if value is None:
        value = ""
    if type is int:
        assert value != "", f"{name} is unset."
        return int(value)
    if type is float:
        assert value != "", f"{name} is unset."
        return float(value)
    if isinstance(type, enum.EnumMeta):
        assert value in type, (
            f'Parameter "{name}" must be one of {[t.value for t in type]}. Found: {value}'
        )
        return type(value)
    opt_type = get_optional_type(type)
    if opt_type:
        return None if value == "" else _param_to_type(name, value, opt_type)
    if isinstance(type, typeof) and issubclass(type, pydantic.BaseModel):
        return type.model_validate_json(value)
    return value


class Op(BaseConfig):
    func: typing.Callable = pydantic.Field(exclude=True)
    categories: list[str]
    name: str
    params: list[Parameter | ParameterGroup]
    inputs: list[Input]
    outputs: list[Output]
    # TODO: Make type an enum with the possible values.
    type: str = "basic"  # The UI to use for this operation.
    color: str = "orange"  # The color of the operation in the UI.
    doc: list | None = None
    # ID is automatically set from the name and categories.
    id: str = pydantic.Field(default=None)

    def __call__(self, *inputs, **params):
        # Convert parameters.
        params = self.convert_params(params)
        res = self.func(*inputs, **params)
        if not isinstance(res, Result):
            # Automatically wrap the result in a Result object, if it isn't already.
            if self.type in [
                "visualization",
                "table_view",
                "graph_creation_view",
                "image",
                "molecule",
            ]:
                # If the operation is a visualization, we use the returned value for display.
                res = Result(display=res)
            else:
                res = Result(output=res)
        return res

    def get_input(self, name: str):
        """Returns the input with the given name."""
        for i in self.inputs:
            if i.name == name:
                return i
        raise ValueError(f"Input {name} not found in operation {self.name}.")

    def get_output(self, name: str):
        """Returns the output with the given name."""
        for o in self.outputs:
            if o.name == name:
                return o
        raise ValueError(f"Output {name} not found in operation {self.name}.")

    def convert_params(self, params: dict[str, typing.Any]):
        """Returns the parameters converted to the expected type."""
        res = dict(params)
        for p in self.params:
            if p.name in params:
                res[p.name] = _param_to_type(p.name, params[p.name], p.type)
        return res

    @pydantic.model_validator(mode="after")
    def compute_id(self):
        self.id = " > ".join(self.categories + [self.name])
        return self


def op(
    env: str | None,
    *names: str,
    view: str = "basic",
    outputs: list[str] | None = None,
    params: list[Parameter] | None = None,
    slow: bool = False,
    color: str | None = None,
    cache: bool | None = None,
    dir: str = "left-to-right",
):
    """
    Decorator for defining an operation.

    Parameters:
        env: The environment (workspace type) to which the operation belongs. If None, the operation
             is not registered in any catalog.
        names: The list of categories this operation belongs to, followed by the name of the operation.
        view: How the operation will be displayed in the UI. One of "basic", "visualization",
              "table_view", "graph_creation_view", "image", "molecule", "matplotlib", "service".
        outputs: A list of output names. If not provided, defaults to ["output"] for "basic" view.
        params: Normally the parameters are taken from the function signature.
                Use "params" to override this.
        slow: If True, the operation results will be cached.
              If the function is not async, it will be run in a separate thread.
        color: The color of the operation in the UI. Defaults to "orange".
        cache: Set to False to disable caching for a slow operation.
               You may need this for slow operations with parameters/outputs that can't be serialized.
        dir: Sets the default input and output positions. The default is "left-to-right", meaning
             inputs are on the left and outputs are on the right. Other options are "right-to-left",
             "top-to-bottom", and "bottom-to-top".
    """
    [*categories, name] = names

    def decorator(func):
        doc = parse_doc(func)
        sig = inspect.signature(func)
        _view = view
        if view == "matplotlib":
            _view = "image"
            func = matplotlib_to_image(func)
        if slow:
            func = make_async(func)
            if cache is not False:
                func = cached(func)
        # Positional arguments are inputs.
        ipos, opos = Position.from_dir(dir)
        inputs = [
            Input(name=name, type=param.annotation, position=ipos)
            for name, param in sig.parameters.items()
            if param.kind not in (param.KEYWORD_ONLY, param.VAR_KEYWORD)
        ]
        _params = []
        for n, param in sig.parameters.items():
            if param.kind == param.KEYWORD_ONLY and not n.startswith("_"):
                _params.append(Parameter.basic(n, param.default, param.annotation))
        if params:
            _params.extend(params)
        if outputs is not None:
            _outputs = [Output(name=name, type=None, position=opos) for name in outputs]
        else:
            _outputs = [Output(name="output", type=None, position=opos)] if view == "basic" else []
        op = Op(
            func=func,
            doc=doc,
            name=name,
            categories=categories,
            params=_params,
            inputs=inputs,
            outputs=_outputs,
            type=_view,
            color=color or "orange",
        )
        if env is not None:
            CATALOGS.setdefault(env, {})
            CATALOGS[env][op.id] = op
        func.__op__ = op
        return func

    return decorator


def matplotlib_to_image(func):
    """Decorator for converting a matplotlib figure to an image."""
    import base64
    import io

    import matplotlib.pyplot as plt
    import matplotlib

    # Make sure we use the non-interactive backend.
    matplotlib.use("agg")

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)
        buf = io.BytesIO()
        plt.savefig(buf, format="png", dpi=300)
        plt.close()
        buf.seek(0)
        image_base64 = base64.b64encode(buf.read()).decode("utf-8")
        return f"data:image/png;base64,{image_base64}"

    return wrapper


def input_position(**positions):
    """
    Decorator for specifying unusual positions for the inputs.

    Example usage:

        @input_position(a="bottom", b="bottom")
        @op("test", "maybe add")
        def maybe_add(a: list[int], b: list[int] | None = None):
            return [a + b for a, b in zip(a, b)] if b else a
    """

    def decorator(func):
        op = func.__op__
        for k, v in positions.items():
            op.get_input(k).position = Position(v)
        return func

    return decorator


def output_position(**positions):
    """Decorator for specifying unusual positions for the outputs.

    Example usage:

        @output_position(output="top")
        @op("test", "maybe add")
        def maybe_add(a: list[int], b: list[int] | None = None):
            return [a + b for a, b in zip(a, b)] if b else a
    """

    def decorator(func):
        op = func.__op__
        for k, v in positions.items():
            op.get_output(k).position = Position(v)
        return func

    return decorator


def no_op(*args, **kwargs):
    if args:
        return args[0]
    return None


def register_passive_op(
    env: str, *names: str, inputs=[], outputs=["output"], params=[], dir="left-to-right", **kwargs
):
    """A passive operation has no associated code."""
    ipos, opos = Position.from_dir(dir)
    [*categories, name] = names
    op = Op(
        func=no_op,
        name=name,
        categories=categories,
        params=params,
        inputs=[
            Input(name=i, type=None, position=ipos) if isinstance(i, str) else i for i in inputs
        ],
        outputs=[
            Output(name=o, type=None, position=opos) if isinstance(o, str) else o for o in outputs
        ],
        **kwargs,
    )
    CATALOGS.setdefault(env, {})
    CATALOGS[env][op.id] = op
    return op


COMMENT_OP = Op(
    func=no_op,
    name="Comment",
    categories=[],
    params=[Parameter.basic("text", "", LongStr)],
    inputs=[],
    outputs=[],
    type="comment",
)


def register_executor(env: str):
    """Decorator for registering an executor.

    The executor is a function that takes a workspace and executes the operations in it.
    When it starts executing an operation, it should call `node.publish_started()` to indicate
    the status on the UI. When the execution is finished, it should call `node.publish_result()`.
    This will update the UI with the result of the operation.
    """

    def decorator(func: typing.Callable[[workspace.Workspace], typing.Any]):
        EXECUTORS[env] = func
        return func

    return decorator


def op_registration(env: str, *categories: str, **kwargs):
    """Returns a decorator that can be used for registering functions as operations."""
    return functools.partial(op, env, *categories, **kwargs)


def passive_op_registration(env: str, *categories: str, **kwargs):
    """Returns a function that can be used to register operations without associated code."""
    return functools.partial(register_passive_op, env, *categories, **kwargs)


def make_async(func):
    """Decorator for slow, blocking operations. Turns them into separate threads."""

    if asyncio.iscoroutinefunction(func):
        # If the function is already a coroutine, return it as is.
        return func

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        return await asyncio.to_thread(func, *args, **kwargs)

    return wrapper


CATALOGS_SNAPSHOTS: dict[str, Catalogs] = {}


def save_catalogs(snapshot_name: str):
    CATALOGS_SNAPSHOTS[snapshot_name] = {k: dict(v) for k, v in CATALOGS.items()}


def load_catalogs(snapshot_name: str):
    global CATALOGS
    snap = CATALOGS_SNAPSHOTS[snapshot_name]
    CATALOGS = {k: dict(v) for k, v in snap.items()}


# Generally the same as the data directory, but it can be overridden.
user_script_root = pathlib.Path()


def load_user_scripts(workspace: str):
    """Reloads the *.py in the workspace's directory and higher-level directories."""
    if "plugins loaded" in CATALOGS_SNAPSHOTS:
        load_catalogs("plugins loaded")
    if not user_script_root:
        return
    path = user_script_root / workspace
    assert path.is_relative_to(user_script_root), f"Path '{path}' is invalid"
    for p in path.parents:
        req = p / "requirements.txt"
        if req.exists():
            try:
                install_requirements(req)
            except Exception:
                traceback.print_exc()
        for f in p.glob("*.py"):
            try:
                run_user_script(f)
            except Exception:
                traceback.print_exc()
        if p == user_script_root:
            break


def install_requirements(req: pathlib.Path):
    cmd = ["uv", "pip", "install", "-q", "-r", str(req)]
    subprocess.check_call(cmd)


def run_user_script(script_path: pathlib.Path):
    spec = importlib.util.spec_from_file_location(script_path.stem, str(script_path))
    assert spec
    module = importlib.util.module_from_spec(spec)
    assert spec.loader
    spec.loader.exec_module(module)


@functools.cache
def parse_doc(func) -> list | None:
    """Griffe is an optional dependency. When available, we return the parsed docstring."""
    doc = func.__doc__
    try:
        import griffe
    except ImportError:
        return None
    if doc is None:
        return None
    griffe.logger.setLevel("ERROR")
    ds = griffe.Docstring(doc, parent=_get_griffe_function(func))
    if "----" in doc:
        ds = ds.parse("numpy")
    else:
        ds = ds.parse("google")
    return json.loads(json.dumps(ds, cls=griffe.JSONEncoder))


def _get_griffe_function(func):
    """Returns a griffe.Function object based on the signature of the function."""
    import griffe

    sig = inspect.signature(func)
    parameters = []
    for name, param in sig.parameters.items():
        if param.annotation is inspect.Parameter.empty:
            annotation = None
        else:
            annotation = getattr(param.annotation, "__name__", str(param.annotation))
        parameters.append(
            griffe.Parameter(
                name,
                annotation=annotation,
                kind=griffe.ParameterKind.keyword_only
                if param.kind == inspect.Parameter.KEYWORD_ONLY
                else griffe.ParameterKind.positional_or_keyword,
            )
        )
    return griffe.Function(
        func.__name__,
        parameters=griffe.Parameters(*parameters),
        returns=str(sig.return_annotation),
    )


def detect_plugins():
    """Imports all installed LynxKite plugins."""
    plugins = {}
    for _, name, _ in pkgutil.iter_modules():
        if (
            name.startswith("lynxkite_")
            and name != "lynxkite_app"
            and name != "lynxkite_core"
            and name != "lynxkite_mcp"
        ):
            plugins[name] = importlib.import_module(name)
    if not plugins:
        print("No LynxKite plugins found. Be sure to install some!")
    return plugins



"""
A LynxKite executor that assumes most operations operate on their input one by one.
"""

import contextlib
import urllib.parse
from .. import ops
from .. import workspace
import traceback
import inspect
import typing


class Context(ops.BaseConfig):
    """Passed to operation functions as "_ctx" if they have such a parameter.

    Attributes:
        node: The workspace node that this context is associated with.
        last_result: The last result produced by the operation.
            This can be used to incrementally build a result, when the operation
            is executed for multiple items.
    """

    node: workspace.WorkspaceNode
    last_result: typing.Any = None


def _df_to_list(df):
    return df.to_dict(orient="records")


def _has_ctx(op):
    sig = inspect.signature(op.func)
    return "_ctx" in sig.parameters


def register(env: str, cache: bool = True):
    """Registers the one-by-one executor.

    Usage:

        from lynxkite_core.executors import one_by_one
        one_by_one.register("My Environment")
    """
    ops.EXECUTORS[env] = lambda ws, ctx: _execute(ws, ops.CATALOGS[env], ctx)


def _get_stages(ws, catalog: ops.Catalog):
    """Inputs on top/bottom are batch inputs. We decompose the graph into a DAG of components along these edges."""
    nodes = {n.id: n for n in ws.nodes}
    batch_inputs = {}
    inputs = {}
    # For each edge in the workspacce, we record the inputs (sources)
    # required for each node (target).
    for edge in ws.edges:
        inputs.setdefault(edge.target, []).append(edge.source)
        node = nodes[edge.target]
        op = catalog[node.data.op_id]
        if op.get_input(edge.targetHandle).position.is_vertical():
            batch_inputs.setdefault(edge.target, []).append(edge.source)
    stages = []
    for bt, bss in batch_inputs.items():
        upstream = set(bss)
        new = set(bss)
        while new:
            n = new.pop()
            for i in inputs.get(n, []):
                if i not in upstream:
                    upstream.add(i)
                    new.add(i)
        stages.append(upstream)
    stages.sort(key=lambda s: len(s))
    stages.append(set(nodes))
    return stages


async def _await_if_needed(obj):
    if inspect.isawaitable(obj):
        return await obj
    return obj


async def _execute(
    ws: workspace.Workspace,
    catalog: ops.Catalog,
    ctx: workspace.WorkspaceExecutionContext | None = None,
):
    nodes = {n.id: n for n in ws.nodes}
    contexts = {n.id: Context(node=n) for n in ws.nodes}
    edges = {n.id: [] for n in ws.nodes}
    for e in ws.edges:
        edges[e.source].append(e)
    tasks = {}
    NO_INPUT = object()  # Marker for initial tasks.
    for node in ws.nodes:
        op = catalog.get(node.data.op_id)
        if op is None:
            node.publish_error(f'Operation "{node.data.op_id}" not found.')
            continue
        node.publish_error(None)
        # Start tasks for nodes that have no non-batch inputs.
        if all([i.position.is_vertical() for i in op.inputs]):
            tasks[node.id] = [NO_INPUT]
    batch_inputs = {}
    # Run the rest until we run out of tasks.
    stages = _get_stages(ws, catalog)
    for stage in stages:
        next_stage = {}
        while tasks:
            n, ts = tasks.popitem()
            if n not in stage:
                next_stage.setdefault(n, []).extend(ts)
                continue
            node = nodes[n]
            op = catalog[node.data.op_id]
            params = {**node.data.params}
            if _has_ctx(op):
                params["_ctx"] = contexts[node.id]
            results = []
            node.publish_started()
            for task in ts:
                try:
                    inputs = []
                    missing = []
                    for i in op.inputs:
                        if i.position.is_vertical():
                            if (n, i.name) in batch_inputs:
                                inputs.append(batch_inputs[(n, i.name)])
                            else:
                                opt_type = ops.get_optional_type(i.type)
                                if opt_type is not None:
                                    inputs.append(None)
                                else:
                                    missing.append(i.name)
                        else:
                            inputs.append(task)
                    if missing:
                        node.publish_error(f"Missing input: {', '.join(missing)}")
                        break
                    result = op(*inputs, **params)
                    output = await _await_if_needed(result.output)
                except Exception as e:
                    traceback.print_exc()
                    node.publish_error(e)
                    break
                contexts[node.id].last_result = output
                # Returned lists and DataFrames are considered multiple tasks.
                if hasattr(output, "to_dict"):
                    output = _df_to_list(output)
                elif not isinstance(output, list):
                    output = [output]
                results.extend(output)
            else:  # Finished all tasks without errors.
                if op.type == "gradio" and ctx and ctx.app:
                    url = f"/api/lynxkite_graph_analytics/{ws.path}/{node.id}"
                    await mount_gradio(ctx.app, result.output, url)
                    result.display = {"backend": urllib.parse.quote(url)}
                    result.output = None
                if result.display:
                    result.display = await _await_if_needed(result.display)
                for edge in edges[node.id]:
                    t = nodes[edge.target]
                    op = catalog[t.data.op_id]
                    if op.get_input(edge.targetHandle).position.is_vertical():
                        batch_inputs.setdefault((edge.target, edge.targetHandle), []).extend(
                            results
                        )
                    else:
                        tasks.setdefault(edge.target, []).extend(results)
                node.publish_result(result)
        tasks = next_stage
    return contexts


class _ProxyApp:
    def __init__(self, app):
        self._app = app
        self.router = self

    @contextlib.asynccontextmanager
    async def lifespan_context(self, app):
        yield

    def mount(self, path, gradio_app):
        import starlette.routing  # ty: ignore[unresolved-import]

        router = self._app.router
        route = starlette.routing.Mount(path, gradio_app)
        # Overwrite existing route if it exists.
        for i, r in enumerate(router.routes):
            if r.path == path:
                router.routes[i] = route
                break
        else:
            router.routes.insert(0, route)


async def mount_gradio(app, gradio_app, path: str):
    """Mounts a Gradio app onto a Starlette/FastAPI app at the given path."""
    import gradio as gr  # ty: ignore[unresolved-import]

    app = _ProxyApp(app)
    gr.mount_gradio_app(app, gradio_app, path=path)
    # Trigger Gradio lifetime hooks.
    async with app.lifespan_context(app):
        pass



"""A LynxKite executor that simply passes the output of one box to the other."""

import os
from .. import ops
from .. import workspace
import traceback
import inspect
import graphlib


def register(env: str):
    """Registers the simple executor.

    Usage:

        from lynxkite_core.executors import simple
        simple.register("My Environment")
    """
    ops.EXECUTORS[env] = lambda ws, _ctx: execute(ws, ops.CATALOGS[env])


async def await_if_needed(obj):
    if inspect.isawaitable(obj):
        return await obj
    return obj


async def execute(ws: workspace.Workspace, catalog: ops.Catalog):
    nodes = {n.id: n for n in ws.nodes}
    dependencies = {n: [] for n in nodes}
    in_edges = {n: {} for n in nodes}
    for e in ws.edges:
        dependencies[e.target].append(e.source)
        assert e.targetHandle not in in_edges[e.target], f"Duplicate input for {e.target}"
        in_edges[e.target][e.targetHandle] = e.source, e.sourceHandle
    outputs = {}
    ts = graphlib.TopologicalSorter(dependencies)
    for node_id in ts.static_order():
        node = nodes[node_id]
        op = catalog[node.data.op_id]
        params = {**node.data.params}
        node.publish_started()
        try:
            inputs = []
            missing = []
            for i in op.inputs:
                edges = in_edges[node_id]
                if i.name in edges and edges[i.name] in outputs:
                    inputs.append(outputs[edges[i.name]])
                else:
                    opt_type = ops.get_optional_type(i.type)
                    if opt_type is not None:
                        inputs.append(None)
                    else:
                        missing.append(i.name)
            if missing:
                node.publish_error(f"Missing input: {', '.join(missing)}")
                continue
            result = op(*inputs, **params)
            result.output = await await_if_needed(result.output)
            result.display = await await_if_needed(result.display)
            if len(op.outputs) == 1:
                [output] = op.outputs
                outputs[node_id, output.name] = result.output
            elif len(op.outputs) > 1:
                assert type(result.output) is dict, "An op with multiple outputs must return a dict"
                for output in op.outputs:
                    outputs[node_id, output.name] = result.output[output.name]
            node.publish_result(result)
        except Exception as e:
            if not os.environ.get("LYNXKITE_SUPPRESS_OP_ERRORS"):
                traceback.print_exc()
            node.publish_error(e)
    return outputs



"""For working with LynxKite workspaces."""

import json
import pathlib
from typing import Any, Optional, TYPE_CHECKING
import dataclasses
import enum
import os
import pydantic
import tempfile
from . import ops

if TYPE_CHECKING:
    import pycrdt
    import fastapi
    from lynxkite_core import ops


class BaseConfig(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        extra="allow",
    )


class Position(BaseConfig):
    x: float
    y: float


class NodeStatus(enum.StrEnum):
    planned = "planned"
    active = "active"
    done = "done"


class WorkspaceNodeData(BaseConfig):
    title: str
    op_id: str
    params: dict
    display: Optional[Any] = None
    input_metadata: Optional[list[dict]] = None
    error: Optional[str] = None
    collapsed: Optional[bool] = None
    status: NodeStatus = NodeStatus.done
    meta: Optional["ops.Op"] = None

    @pydantic.model_validator(mode="before")
    @classmethod
    def fill_op_id_if_missing(cls, data: dict) -> dict:
        """Compatibility with old workspaces that don't have op_id."""
        if "op_id" not in data:
            data["op_id"] = data["title"]
        return data

    @pydantic.model_validator(mode="before")
    @classmethod
    def ignore_meta(cls, data: dict) -> dict:
        """Metadata is never loaded. We will use fresh metadata."""
        data["meta"] = None
        return data


class WorkspaceNode(BaseConfig):
    # Most of these fields are shared with ReactFlow.
    id: str
    type: str
    data: WorkspaceNodeData
    position: Position
    width: Optional[float] = None
    height: Optional[float] = None
    _crdt: Optional["pycrdt.Map"] = None

    def publish_started(self):
        """Notifies the frontend that work has started on this node."""
        self.data.error = None
        self.data.status = NodeStatus.active
        if self._crdt and "data" in self._crdt:
            with self._crdt.doc.transaction():
                self._crdt["data"]["error"] = None
                self._crdt["data"]["status"] = NodeStatus.active

    def publish_result(self, result: ops.Result):
        """Sends the result to the frontend. Call this in an executor when the result is available."""
        self.data.display = result.display
        self.data.input_metadata = result.input_metadata
        self.data.error = result.error
        self.data.status = NodeStatus.done
        if self._crdt and "data" in self._crdt:
            with self._crdt.doc.transaction():
                try:
                    self._crdt["data"]["status"] = NodeStatus.done
                    self._crdt["data"]["display"] = self.data.display
                    self._crdt["data"]["input_metadata"] = self.data.input_metadata
                    self._crdt["data"]["error"] = self.data.error
                except Exception as e:
                    self._crdt["data"]["error"] = str(e)
                    raise e

    def publish_error(self, error: Exception | str | None):
        """Can be called with None to clear the error state."""
        result = ops.Result(error=str(error) if error else None)
        self.publish_result(result)


class WorkspaceEdge(BaseConfig):
    id: str
    source: str
    target: str
    sourceHandle: str
    targetHandle: str


@dataclasses.dataclass
class WorkspaceExecutionContext:
    """Context passed to ops during execution."""

    app: "fastapi.FastAPI | None"


class Workspace(BaseConfig):
    """A workspace is a representation of a computational graph that consists of nodes and edges.

    Each node represents an operation or task, and the edges represent the flow of data between
    the nodes. Each workspace is associated with an environment, which determines the operations
    that can be performed in the workspace and the execution method for the operations.
    """

    env: str = ""
    nodes: list[WorkspaceNode] = dataclasses.field(default_factory=list)
    edges: list[WorkspaceEdge] = dataclasses.field(default_factory=list)
    path: Optional[str] = None
    _crdt: Optional["pycrdt.Map"] = None

    def normalize(self):
        if self.env not in ops.CATALOGS:
            return self
        catalog = ops.CATALOGS[self.env]
        _ops = {n.id: catalog[n.data.op_id] for n in self.nodes if n.data.op_id in catalog}
        valid_targets = set()
        valid_sources = set()
        for n in self.nodes:
            if n.id in _ops:
                for h in _ops[n.id].inputs:
                    valid_targets.add((n.id, h.name))
                for h in _ops[n.id].outputs:
                    valid_sources.add((n.id, h.name))
        self.edges = [
            edge
            for edge in self.edges
            if (edge.source, edge.sourceHandle) in valid_sources
            and (edge.target, edge.targetHandle) in valid_targets
        ]

    def has_executor(self):
        return self.env in ops.EXECUTORS

    async def execute(self, ctx: WorkspaceExecutionContext | None = None):
        return await ops.EXECUTORS[self.env](self, ctx)

    def model_dump_json(self) -> str:
        """Returns the workspace as JSON."""
        # Pydantic can't sort the keys. TODO: Keep an eye on https://github.com/pydantic/pydantic-core/pull/1637.
        j = self.model_dump()
        if "path" in j:
            del j["path"]
        j = json.dumps(j, indent=2, sort_keys=True) + "\n"
        return j

    def save(self, path: str | pathlib.Path):
        """Persist the workspace to a local file in JSON format."""
        path = str(path)
        j = self.model_dump_json()
        dirname, basename = os.path.split(path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)
        # Create temp file in the same directory to make sure it's on the same filesystem.
        with tempfile.NamedTemporaryFile(
            "w", encoding="utf-8", prefix=f".{basename}.", dir=dirname, delete=False
        ) as f:
            temp_name = f.name
            f.write(j)
        os.replace(temp_name, path)

    @staticmethod
    def load(path: str | pathlib.Path) -> "Workspace":
        """Load a workspace from a file.

        After loading the workspace, the metadata of the workspace is updated.

        Args:
            path: The path to the file to load the workspace from.

        Returns:
            Workspace: The loaded workspace object, with updated metadata.
        """
        path = str(path)
        with open(path, encoding="utf-8") as f:
            j = f.read()
        ws = Workspace.model_validate_json(j)
        # Metadata is added after loading. This way code changes take effect on old boxes too.
        ws.update_metadata()
        ws.path = path
        return ws

    def update_metadata(self):
        """Update the metadata of this workspace.

        The metadata is the information about the operations that the nodes in the workspace represent,
        like the parameters and their possible values.
        This information comes from the catalog of operations for the environment of the workspace.
        """
        if self.env not in ops.CATALOGS:
            return self
        catalog = ops.CATALOGS[self.env]
        for node in self.nodes:
            data = node.data
            op = catalog.get(data.op_id)
            if op:
                if getattr(data, "meta", None) != op:
                    data.meta = op
                    # If the node is connected to a CRDT, update that too.
                    if node._crdt:
                        node._crdt["data"]["meta"] = op.model_dump()
                if node.type != op.type:
                    node.type = op.type
                    if node._crdt:
                        node._crdt["type"] = op.type
                if data.error == "Unknown operation.":
                    data.error = None
                    if node._crdt:
                        node._crdt["data"]["error"] = None
            else:
                data.error = "Unknown operation."
                if node._crdt:
                    node._crdt["data"]["meta"] = {}
                    node._crdt["data"]["error"] = "Unknown operation."

    def connect_crdt(self, ws_crdt: "pycrdt.Map"):
        import pycrdt

        self._crdt = ws_crdt
        with ws_crdt.doc.transaction():
            for nc, np in zip(ws_crdt["nodes"], self.nodes):
                if "data" not in nc:
                    nc["data"] = pycrdt.Map()
                np._crdt = nc

    def add_node(self, func=None, **kwargs):
        """For convenience in e.g. tests."""
        random_string = os.urandom(4).hex()
        if func:
            kwargs["type"] = func.__op__.type
            kwargs["data"] = WorkspaceNodeData(
                title=func.__op__.name, op_id=func.__op__.id, params={}
            )
        elif "title" in kwargs:
            kwargs["data"] = WorkspaceNodeData(
                title=kwargs["title"],
                op_id=kwargs["title"],
                params=kwargs.get("params", {}),
            )
        kwargs.setdefault("type", "basic")
        kwargs.setdefault("id", f"{kwargs['data'].title} {random_string}")
        kwargs.setdefault("position", Position(x=0, y=0))
        kwargs.setdefault("width", 100)
        kwargs.setdefault("height", 100)
        node = WorkspaceNode(**kwargs)  # ty: ignore[missing-argument]
        self.nodes.append(node)
        return node

    def add_edge(
        self,
        source: WorkspaceNode,
        sourceHandle: str,
        target: WorkspaceNode,
        targetHandle: str,
    ):
        """For convenience in e.g. tests."""
        edge = WorkspaceEdge(
            id=f"{source.id} {sourceHandle} to {target.id} {targetHandle}",
            source=source.id,
            target=target.id,
            sourceHandle=sourceHandle,
            targetHandle=targetHandle,
        )
        self.edges.append(edge)
        return edge
